//
// GameX (Release 4 - SoundX)
// Sound Class Code 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 


// DirectSound
#ifndef DIRECTSOUND_VERSION 
#define DIRECTSOUND_VERSION 0x0700 // specify version 7.0 of DirectSound
//#include <dsound.h>
#endif

// DirectMusic
#ifndef DIRECTMUSIC_VERSION
#define DIRECTMUSIC_VERSION 0x0700 // specify version 7.0 of DirectMusic
//#include <dmksctrl.h>
#include <dmusici.h>
//#include <dmusicc.h>
#include <dmusicf.h>
#endif

// DirectShow
#ifndef DIRECTSHOW_VERSION
#define DIRECTSHOW_VERSION 0x0700 // specify version 7.0 of DirectShow
#include <strmif.h>
#include <control.h>
#include <uuids.h>
//#include <evcode.h>
//#include <amvideo.h>
#endif


#include "gamex-sound.h"
#include "gamex-win-dx.h"
#include "gamex-defines.h"

#include <direct.h>


// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif


SoundX::SoundX()
{
	z_samples_per_sec = 0;
	z_bits_per_sample = 0;
	z_length = 0;
	z_num_bytes = 0;
	z_num_samples = 0;	
	z_data = NULL;
	z_dx_refresh = true;
	z_ds_sound_buffer = NULL;
	z_filename[0] = '\0';
}

SoundX::~SoundX()
{
	Unload();
}

static bool inverseFormat = false;

bool SoundX::ReadName (FILE* wav, const char *match_name)
{
	char buf[5];
	fread(buf, 4, 1, wav);

	if(inverseFormat)
	{
		for(int i = 0 ; i < 4 ; i++)
			buf[i] = ~buf[i];
	}

	return (strncmp(buf, match_name, 4) == 0);
}

bool SoundX::ReadSize (FILE* wav, unsigned int &size)
{
	size = 0;
	fread(&size, 4, 1, wav);

	if(inverseFormat)
		size = ~size;

	return true;
}

bool SoundX::LoadWav (const char *filename)
{
	GameX.z_InitDirectSound();

	unsigned short twobytes;
	unsigned long fourbytes;
	unsigned short format;
	unsigned int size;
	int source_bps;
	char buf[100];
	ZeroMemory(buf,100);
	int n;

	inverseFormat = false;
	char newFilename [MAX_PATH];
	if(!debug.DoesFileExist(filename))
	{
		strcpy(newFilename, filename);
		char* dot = strrchr(newFilename, '.');
		if(dot)
		{
			strcpy(dot, ".gxs");
			if(debug.DoesFileExist(newFilename))
			{
				filename = (char*)newFilename;
				inverseFormat = true;
			}
		}
	}
	
	if (strstr(filename, ".gxs"))
		inverseFormat = true;
	
	FILE* wav = fopen(filename, "rb");
	if(wav)
	{
		if(inverseFormat)
		{
			fread(&buf, 64, 1, wav);
			ZeroMemory(buf,100);
		}

		if(ReadName(wav, "RIFF") == false) {InvalidWav(); return false;}
		if(ReadSize(wav, size) == false)	 {InvalidWav(); return false;}
		if(ReadName(wav, "WAVE") == false) {InvalidWav(); return false;}
		if(ReadName(wav, "fmt ") == false) {InvalidWav(); return false;}
		if(ReadSize(wav, size) == false)	 {InvalidWav(); return false;}

		// Read WAV Format Header (THIS READS 16 BYTES)
		fread(&format, 2, 1, wav);										// WAV Format Tag (1 = PCM) (2 = MS ADPCM)
		if(inverseFormat)
			format = ~format;
		switch(format)
		{
		case 1:
		{
			fread(&z_num_channels, 2, 1, wav);							// WAV Number of Channels
			fread(&z_samples_per_sec, 4, 1, wav);						// WAV Samples per Second
			fread(&fourbytes, 4, 1, wav);								// WAV Average Bytes per Second (IGNORED)
			fread(&twobytes, 2, 1, wav);								// WAV Block Align (IGNORED)
			fread(&z_bits_per_sample, 2, 1, wav);						// WAV Bits Per Sample

			if(inverseFormat)
			{
				z_num_channels = ~z_num_channels;
				z_samples_per_sec = ~z_samples_per_sec;
				z_bits_per_sample = ~z_bits_per_sample;
			}

			source_bps = z_bits_per_sample;
			z_bits_per_sample = min(16,source_bps);
			fread(buf, size-16, 1, wav); // ignored?
			if(inverseFormat) // (in case it isn't ignored)
				FOR(j, size-16)
					buf[j] = ~buf[j];
			while(!feof(wav) && !ferror(wav))
			{
				fread(buf, 1, 1, wav);
				if(feof(wav) || ferror(wav))
					break;
				if(inverseFormat)
					buf[0] = ~buf[0];
				if(buf[0] == 'd')
				{
					fread(buf+1, 3, 1, wav);
					if(inverseFormat)
						FOR(j, 3)
							buf[j+1] = ~buf[j+1];
				}
				else if(buf[0] == 'f')
				{
					fread(buf+1, 3, 1, wav);
					if(inverseFormat)
						FOR(j, 3)
							buf[j+1] = ~buf[j+1];
				}
				else
					continue;
				if(strncmp(buf, "data", 4) == 0)
				{
					if(ReadSize(wav, size) == false)	{InvalidWav(); return false;}
					// Read WAV Data
					z_num_bytes = size;														// Load number of bytes in z_data (from file)				
					z_num_samples = z_num_bytes / ((z_bits_per_sample/8) * z_num_channels);	// Calculate number of samples
					z_length = (float) z_num_samples / (float) z_samples_per_sec;			// Calculate z_length of sound (in seconds)
					SAFE_DELETE_ARRAY(z_data);
					
					z_data = new char[z_num_bytes];
					ZeroMemory(z_data, z_num_bytes);

					int source_byte_ps = source_bps/8;
					int dest_byte_ps = z_bits_per_sample/8;
					int skip_channels = (source_byte_ps-dest_byte_ps);
					int read_channels = dest_byte_ps;

					for(n = 0; n < z_num_bytes; n += dest_byte_ps)
					{
						if(skip_channels)
						{
							char skip [8];
							fread(&skip, skip_channels, 1, wav);
						}
						fread(z_data + n, read_channels, 1, wav);
						if(inverseFormat)
							FOR(j, read_channels)
								z_data[n+j] = ~z_data[n+j];
					}
				}
				else if(strncmp (buf, "fact", 4) == 0)
				{
					fread(&fourbytes, 4, 1, wav);
					fread(&fourbytes, 4, 1, wav);
				}
			}

			GameX.z_RegisterSoundX(this);
		}	break;
		default:
			assertMsg2(0, "The WAV \"%s\" was not saved in PCM format. Please re-save it in any standard PCM format so GameX can load it.", filename);
			fclose(wav);
			return false;
		}
				
		z_dx_refresh = true;
		fclose(wav);
	}
	else
	{
		assertMsg2(0, ".WAV file not found: %s\n\nMake sure names of WAV files are correct and somewhere in the game's directory.", filename);
		return false;
	}

	sprintf(z_filename, filename);

	if(GameX.z_win_log_debug)
	{
		int z;
		char debug_message [MAX_PATH];
		for(z = 0 ; filename[z] ; z++)
		{
			debug_message[z] = filename[z];
		}
		debug_message[z++] = ':';
		while(z < 48)
		{
			debug_message[z++] = ' ';
		}
		debug_message[z++] = (char)0;
//		sprintf(debug_message,"%s:", filename);
		debug.Output(debug_message); 

		char sizeStr [8];
		sprintf(sizeStr, "%05d", (z_num_bytes+1023)/1024);
		for(z = 0 ; sizeStr[z] == '0' ; z++) sizeStr[z] = ' ';

		debug.Output("%s KB, %dx%dx%d\n",sizeStr,z_num_samples,z_num_channels,z_bits_per_sample);
	}
	GameX.z_MemoryRecordChange(z_num_bytes);


	return true;
}

void SoundX::Unload ()
{
	if(GameX.z_GetDS()) // if we haven't already released all of DirectSound
	{
		if(GameX.IsSoundPlaying(this))
		{
			GameX.StopSound(this);
			if(debug.IsRunningDebug())
				GameX.WarningDialog("A SoundX object was deleted while it was still playing.\nIf you meant to delete it, stop it from playing first.");
		}
		if(z_ds_sound_buffer != NULL)
		{
			z_ds_sound_buffer->Release(), z_ds_sound_buffer = NULL;
			GameX.z_MemoryRecordChange(-z_num_bytes);
		}
	}
	if(z_data != NULL)
	{
		SAFE_DELETE_ARRAY(z_data);
		GameX.z_MemoryRecordChange(-z_num_bytes);
	}
	z_num_bytes = 0;
	GameX.z_UnRegisterSoundX(this);
}

void SoundX::InvalidWav ()
{
	char disp[500];
	sprintf (disp, ".WAV Format Invalid: \n\nInvalid WAV Format. Not all types of WAV files are supported. Try saving your WAV file with a different program, or with 'extra wav information' turned off."); 
	if(debug.IsRunningDebug())
		debug.AssertBox(disp);
	else
		debug.Output(disp);
}
/*
// looks buggy to me, and nothing uses it => commented it out
void SoundX::CutLength (float new_length)
{
	int new_samples, new_bytes;
	char* new_data;

	if(z_data != NULL && z_length > new_length)
	{
		new_samples = (int) (z_samples_per_sec * new_length);
		new_bytes = (z_bits_per_sample/8) * z_num_channels * new_samples;	
		new_data = new char[new_bytes];
		memcpy (new_data, z_data, new_bytes);
		SAFE_DELETE_ARRAY(z_data);
		z_data = new_data;
		SAFE_DELETE_ARRAY(new_data);
		z_num_bytes = new_bytes;
		z_num_samples = new_samples;
		z_length = new_length;
		z_dx_refresh = true;
	}
}
*/
bool SoundX::NeedUpdate () const
{
	return z_dx_refresh;
}


void SoundX::PrepareDSBuffer () const
{
	WAVEFORMATEX format; 
	ZeroMemory(&format,sizeof(format)) ;		
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = z_num_channels;
	format.nSamplesPerSec = z_samples_per_sec;
	format.wBitsPerSample = z_bits_per_sample;
	format.nBlockAlign = (format.wBitsPerSample / 8) * format.nChannels; // /8 converts bits to bytes
	format.nAvgBytesPerSec = format.nSamplesPerSec * format.nBlockAlign;	
	
	DSBUFFERDESC bufferDesc; 
	ZeroMemory(&bufferDesc,sizeof(bufferDesc)) ;  
	bufferDesc.dwSize = sizeof(bufferDesc) ; 
	bufferDesc.dwFlags = DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME; 
	bufferDesc.dwBufferBytes = z_num_samples * format.nBlockAlign;
	bufferDesc.lpwfxFormat = (LPWAVEFORMATEX) &format;

	LPDIRECTSOUND z_ds_main = GameX.z_GetDS();

	if(z_ds_sound_buffer != NULL)
	{
		z_ds_sound_buffer->Release(), z_ds_sound_buffer = NULL;
		GameX.z_MemoryRecordChange(-z_num_bytes);
	}

	HRESULT status = z_ds_main->CreateSoundBuffer(&bufferDesc, &z_ds_sound_buffer, NULL);
	if(FAILED (status))
	{
		switch(status)
		{
 		case DSERR_ALLOCATED: debug.Output("DSERR_ALLOCATED"); break;
		case DSERR_BADFORMAT: debug.Output("DSERR_BADFORMAT"); break;
		case DSERR_BUFFERTOOSMALL: debug.Output("DSERR_BUFFERTOOSMALL"); break;
		case DSERR_CONTROLUNAVAIL: debug.Output("DSERR_CONTROLUNAVAIL"); break;
		case DSERR_DS8_REQUIRED: debug.Output("DSERR_DS8_REQUIRED"); break;
		case DSERR_INVALIDCALL: debug.Output("DSERR_INVALIDCALL"); break;
		case DSERR_INVALIDPARAM: debug.Output("DSERR_INVALIDPARAM"); break;
		case DSERR_NOAGGREGATION: debug.Output("DSERR_NOAGGREGATION"); break;
		case DSERR_OUTOFMEMORY: debug.Output("DSERR_OUTOFMEMORY"); break;
		case DSERR_UNINITIALIZED: debug.Output("DSERR_UNINITIALIZED"); break;
		case DSERR_UNSUPPORTED: debug.Output("DSERR_UNSUPPORTED"); break;
		}
		GameX.z_ReportProblem (GAMEX_SOUNDCREATE_FAILED) ;
	}
	
	return;
}

#define MY_NAME ((const char*)(*this))

MusicX::MusicX()
{
	z_segment = NULL;
	z_segment_state = NULL;
	z_registered = false;
	z_vol = 0;
	z_playstate = 0;
	dsh_gb = NULL;
	dsh_mc = NULL;
	dsh_mp = NULL;
	dsh_ba = NULL;
	dsh_bv = NULL;
	dsh_me = NULL;
	dsh_vw = NULL;
	Release();
}

MusicX::~MusicX()
{
	if(GameX.z_GetDMP()) { // if we haven't already released all of DirectMusic
		if(GameX.IsMusicPlaying(this))
		{
			GameX.PlayMusic(NULL); // stops music and sets current music to NULL
			if(debug.IsRunningDebug())
				GameX.WarningDialog("A %s object was deleted while it was still playing.\nIf you meant to delete it, stop it first.", MY_NAME);
		}
		Release();
	}
	else GameX.z_UnRegisterMusicX(this);
}

void MusicX::Release ()
{
	if(z_segment != NULL)
	{
		z_segment->SetParam(GUID_Unload, -1, 0, 0,(void*)GameX.z_GetDMP());
		z_segment->Release(), z_segment = NULL;
	}
	if(z_segment_state != NULL) z_segment_state->Release(), z_segment_state = NULL;

	// DirectShow
	if(dsh_gb) 		dsh_gb->Release(),			 dsh_gb = NULL;
	if(dsh_mc) 		dsh_mc->Release(),			 dsh_mc = NULL;
	if(dsh_mp) 		dsh_mp->Release(),			 dsh_mp = NULL;
	if(dsh_ba) 		dsh_ba->Release(),			 dsh_ba = NULL;
	if(dsh_bv) 		dsh_bv->Release(),			 dsh_bv = NULL;
	if(dsh_me) 		dsh_me->Release(),			 dsh_me = NULL;
	if(dsh_vw) 		dsh_vw->Release(),			 dsh_vw = NULL;

	playing_loops_left = 0;
	format = MEDIA_TYPE_NONE;

	GameX.z_UnRegisterMusicX(this);
}

bool MusicX::Load (char * filename)
{
	char * extension = strrchr(filename, '.'); // string of text from the last dot onward
	if(extension && (_stricmp(extension, ".MID") == 0 || _stricmp(extension, ".MIDI") == 0)) // case insensitive match
		return z_LoadWithDirectMusic (filename);
	else
	if(extension && (_stricmp(extension, ".MP3") == 0) || extension && (_stricmp(extension, ".WMA") == 0))
		return z_LoadWithDirectShow (filename);
	else
	{
		assertMsg3(0, "%s: Did not load %s because it does not appear to be in a supported music format (.mid, mp3, or .wma).", MY_NAME, filename);
		return false;
	}
}

bool MovieX::Load (char * filename)
{
	char * extension = strrchr(filename, '.');
	if(extension && (_stricmp(extension, ".MID") == 0 || _stricmp(extension, ".MIDI") == 0
	              || _stricmp(extension, ".MP3") == 0 || _stricmp(extension, ".WMA") == 0))
	{
		assertMsg3(0, "%s: Did not load %s because it appears to be a music file.\nTry using MusicX instead of MovieX to load it.", MY_NAME, filename);
		return false;
	}
	else
	{
		return z_LoadWithDirectShow (filename);
	}
}

bool MusicX::z_LoadWithDirectMusic (char *filename)
{
	// this function loads a midi z_segment
	DMUS_OBJECTDESC obj_desc;

	Release();

	format = MEDIA_TYPE_DMUSIC;

	// get current working directory
	char dir [_MAX_PATH];
	WCHAR w_dir [_MAX_PATH];
	if(_getcwd(dir, _MAX_PATH) == NULL)
		return false;

	MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,dir,-1,w_dir,_MAX_PATH);

	GameX.z_InitDirectMusicLoader();
	IDirectMusicLoader * dml = GameX.z_GetDML();

	// tell the loader were to look for files
	if(dml == NULL || FAILED(dml->SetSearchDirectory(GUID_DirectMusicAllTypes, w_dir, FALSE)))
		return false;

	// convert filename to wide string
	WCHAR w_filename[_MAX_PATH];
	MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,filename,-1,w_filename,_MAX_PATH);

	// setup object description
	ZeroMemory(&obj_desc, sizeof(obj_desc));
	obj_desc.dwSize = sizeof(obj_desc);
	obj_desc.guidClass = CLSID_DirectMusicSegment;
	wcscpy(obj_desc.wszFileName, w_filename);
	obj_desc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME;

	GameX.z_InitDirectMusicPerformance();

	if(FAILED(GameX.z_GetDML()->GetObject(&obj_desc,IID_IDirectMusicSegment, (void**) &z_segment))
	|| FAILED(z_segment->SetParam(GUID_StandardMIDIFile,-1, 0, 0, (void*)GameX.z_GetDMP()))
	|| FAILED(z_segment->SetParam(GUID_Download, -1, 0, 0, (void*)GameX.z_GetDMP())))
	{
		Release();
		return false;
	}

	DMUS_TEMPO_PARAM original_tempo;
	z_segment->GetParam(GUID_TempoParam, 0xFFFF, 0, 0, NULL, &original_tempo);

	if(!z_registered)
	{
		GameX.z_RegisterMusicX(this);
		z_registered = true;
	}

	return true;
}

bool MusicX::z_LoadWithDirectShow (char *filename)
{
	if(!GameX.z_CanPlayMP3()) return false;

	bool success = z_SetUpFilterGraph();
	if(!success) return false;

	format = MEDIA_TYPE_DSHOW;


	// convert filename to wide string
	WCHAR w_filename[_MAX_PATH];
	MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,filename,-1,w_filename,_MAX_PATH);

	HRESULT status;
	try
	{
		status = dsh_gb->RenderFile(w_filename,NULL);
	}
	catch(...)
	{
		status = -1;
	}

	if(FAILED(status))
	{
		assertMsg3(0, "%s: Failed to load the file \"%s\". This may be caused by it having the wrong extension (such as .mp3 vs .wma), or it may simply be an invalid or unsupported media format.", MY_NAME, filename);
		return false;
	}
	else
	{
		z_PrepareLoaded();

		return true;
	}
}

bool MusicX::z_SetUpFilterGraph ()
{
	Release();

	HRESULT status;

	status = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC, IID_IGraphBuilder, (void **)&dsh_gb);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create graph builder.\n", MY_NAME);
		return false;
	}

	status = dsh_gb->QueryInterface(IID_IMediaControl, (void **)&dsh_mc);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create media control.\n", MY_NAME);
		return false;
	}

	status = dsh_gb->QueryInterface(IID_IMediaPosition,(void**)&dsh_mp);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create media position.\n", MY_NAME);
		return false;
	}

	status = dsh_gb->QueryInterface(IID_IMediaEventEx,(void**)&dsh_me);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create media event extender.\n", MY_NAME);
		return false;
	}

	status = dsh_gb->QueryInterface(IID_IBasicAudio,(void**)&dsh_ba);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create basic audio modifier.\n", MY_NAME);
		return false;
	}

	status = dsh_me->SetNotifyWindow((OAHWND)GameX.GetWindow(), WM_DIRECTSHOW_EVENT, 0);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to set music notify callback.\n", MY_NAME);
		return false;
	}

	return true;
}

bool MovieX::z_SetUpFilterGraph ()
{
	bool musicSetup = this->MusicX::z_SetUpFilterGraph();

	HRESULT status = dsh_gb->QueryInterface(IID_IBasicVideo,(void**)&dsh_bv);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create basic video modifier.\n", MY_NAME);
		return false;
	}

	status = dsh_gb->QueryInterface(IID_IVideoWindow,(void**)&dsh_vw);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to create video window modifier.\n", MY_NAME);
		return false;
	}

	//status = CoCreateInstance(CLSID_VideoRenderer, NULL, CLSCTX_INPROC, IID_IDirectDrawVideo, (void **)&dd_video);
	//if(FAILED(status))
	//{
	//	assertMsg2(0, "%s: Failed to create DirectDraw video window interface.\n", MY_NAME);
	//	return false;
	//}

	//LPDIRECTDRAW7 z_dd_main = GameX.z_GetDD();
	//status = dd_video->SetDirectDraw((LPDIRECTDRAW)z_dd_main);
	//if(FAILED(status))
	//{
	//	assertMsg2(0, "%s: Failed to set DirectDraw as video window.\n", MY_NAME);
	//	return false;
	//}

	return musicSetup;
}

bool MusicX::z_PrepareLoaded ()
{
	return true;
}

bool MovieX::z_PrepareLoaded ()
{
	bool musicLoaded = this->MusicX::z_PrepareLoaded();

	HRESULT status = dsh_vw->put_Owner((OAHWND)GameX.GetWindow());
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to put_Owner the video window.\n", MY_NAME);
		return false;
	}

	status = dsh_vw->put_WindowStyle(WS_CHILD | WS_CLIPSIBLINGS);
	if(FAILED(status))
	{
		assertMsg2(0, "%s: Failed to put_WindowStyle the video window.\n", MY_NAME);
		return false;
	}

	// cache the movie's original size
	dsh_vw->get_Width(&z_xres);
	dsh_vw->get_Height(&z_yres);

	// default is to start the movie filling 7/8 of the window along its largest side
	if(z_xres > z_yres)
		SetPosition(GameX.GetWidth() / 2, GameX.GetHeight() / 2, 7.0f/8.0f * (float)GameX.GetWidth() / (float)z_xres);
	else																	
		SetPosition(GameX.GetWidth() / 2, GameX.GetHeight() / 2, 7.0f/8.0f * (float)GameX.GetHeight() / (float)z_yres);

	dsh_vw->put_Visible(true);

	return musicLoaded;
}

void MovieX::SetPosition (int xLeft, int yTop, int xRight, int yBottom)
{
	z_x1 = xLeft;
	z_y1 = yTop;
	z_x2 = xRight;
	z_y2 = yBottom;

	AdjustSize();
}

void MovieX::SetPosition (int xCenter, int yCenter, float scaleFactor)
{
	const long halfWidth = Round(z_xres * 0.5f);
	const long halfHeight = Round(z_yres * 0.5f);
	SetPosition(xCenter - halfWidth, yCenter - halfHeight, xCenter + halfWidth, yCenter + halfHeight);
}

void MovieX::AdjustSize ()
{
	RECT rect;
	GetClientRect (GameX.GetWindow(), &rect);

	const int x1 = z_x1 * GameX.z_win_realx / GameX.z_win_width  + GameX.z_win_xoffset;
	const int y1 = z_y1 * GameX.z_win_realy / GameX.z_win_height + GameX.z_win_yoffset;
	const int x2 = z_x2 * GameX.z_win_realx / GameX.z_win_width  + GameX.z_win_xoffset;
	const int y2 = z_y2 * GameX.z_win_realy / GameX.z_win_height + GameX.z_win_yoffset;

	dsh_vw->SetWindowPosition(x1, y1, x2 - x1, y2 - y1);
}

// sends a curve message (to alter volume or pitch over time)
void MusicX::Curve(long time, int target, int type, int shape, int start)
{
	if(format != MEDIA_TYPE_DMUSIC) return; // not supported for MP3s (must be a DirectShow filter for it though)

	DMUS_CURVE_PMSG *pCurveMsg;
	HRESULT hr;

	if(NULL == GameX.z_GetDMP()) return;
	hr = GameX.z_GetDMP()->AllocPMsg(sizeof(DMUS_CURVE_PMSG), (DMUS_PMSG**) &pCurveMsg);
	if(SUCCEEDED(hr))
	{
		ZeroMemory(pCurveMsg, sizeof(DMUS_CURVE_PMSG));
		pCurveMsg->dwSize = sizeof(DMUS_CURVE_PMSG);
		pCurveMsg->rtTime = 0;
		pCurveMsg->dwFlags = DMUS_PMSGF_REFTIME | DMUS_PMSGF_LOCKTOREFTIME;
		pCurveMsg->dwPChannel = DMUS_PCHANNEL_BROADCAST_PERFORMANCE;
		pCurveMsg->dwType = DMUS_PMSGT_CURVE;
		pCurveMsg->dwGroupID = 0xFFFFFFF;
		pCurveMsg->mtDuration = time;
		pCurveMsg->nEndValue = target;
		pCurveMsg->bCurveShape = shape;
		pCurveMsg->bCCData = 7;
		if(start == -1)
		{
			pCurveMsg->bFlags = DMUS_CURVE_START_FROM_CURRENT;
		}
		else
		{
			pCurveMsg->bFlags = 0;
			pCurveMsg->nStartValue = start;
		}
		pCurveMsg->bType = type ;
		pCurveMsg->wMergeIndex = merge_id++; // curves attempt to merge unless they have identical merge indexes
		hr = GameX.z_GetDMP()->SendPMsg((DMUS_PMSG*) pCurveMsg);
	}
}


